# Qubit teleportation between non-neighboring nodes in a quantum network
# Hermans, Pompili, et al.

import qutip as qt
import numpy as np

parameters = {
    "babysitter_slowdown": 0.4, 
    "epr_ab": {
        "dc_rate": 10,
        "det_win": 15e-9,
        "pdet_1": 1.7e-4 * 2,
        "pdet_2": 2.55e-4 * 2,
        "vis": 0.90,
        "p2e": 0.06,
        "phase_u": 21.0,
        "lde_length": 3.8e-6,
        "depol": 0.0,
        "pdet_psb" : 0.1, # efficiency of the psb detection
    },
    "epr_bc": {
        "dc_rate": 10,
        "det_win": 15e-9,
        "pdet_1": 2.14e-4 * 2,
        "pdet_2": 1.2e-4 * 2,
        "vis": 0.90,
        "p2e": 0.08,
        "phase_u": 12.,
        "lde_length": 5.0e-6,
        "depol": 0.0,
        "pdet_psb" : 0.12, # efficiency of the psb detection 
    },
    "communication_alice" : {
        "ionization" : 0.007, #actual fraction, not percents
        "dephasing" : 0.0, #dephasing at the worst possible point
        "depolarizing" : 0.04, #depolarization at the worst possible point
    },
    "communication_bob" : {
        "dephasing" : 0,#0.022/2., #dephasing at the worst possible point
        "depolarizing" : 0,#0.06/2., #depolarization at the worst possible point
    },
    "communication_charlie" : {
        "dephasing" : 0,#0.018/2., #dephasing at the worst possible point
        "depolarizing" : 0,#0.022/2., #depolarization at the worst possible point
    },
    "memory_bob": {
        "depol":(0.12),
        "alphas": [0.05],
        "N1e": [5300], 
        "power":[1.1],
    },
    "memory_charlie": {
        "depol":(0.14),
    },
    "bob_ro_fidelities": {
	    'communication' : {
	        "0": 0.93,
	        "1": 0.995,
	        },
        'memory' : {
	        "0": 0.99,
	        "1": 0.99,
        }
    },
    "charlie_ro_fidelities": { 
	    'communication' : {
	        "0": 0.92,
	        "1": 0.99,
	        },
        'memory' : {
	        "0": 0.98,
	        "1": 0.98,
        }
    },
    "BSM": {
        "bob" : ["00","10"], #memory, communication
        "charlie" : ["00", "10"], #memory, communication
    },
    
}

def _actual_fidelity(x, y):
    return qt.fidelity(x, y)**2

def _circuit_list_to_propagator(circuit_list):
    propagator = circuit_list[-1]
    for g in reversed(circuit_list[:-1]):
        propagator *= g
    return propagator

def _transform_rho(rho, gate):
    return gate * rho * gate.dag()

I = qt.identity(2)
X = qt.sigmax()
Y = qt.sigmay()
Z = qt.sigmaz()
H = qt.operations.hadamard_transform(1)
II = qt.identity([2, 2])
ZI = qt.tensor(Z, I)
IZ = qt.tensor(I, Z)
XX = qt.tensor(X, X)
YY = qt.tensor(Y, Y)
ZZ = qt.tensor(Z, Z)
IH = qt.tensor(I, H)
IX = qt.tensor(I, X)
IY = qt.tensor(I, Y)
HI = qt.tensor(H,I)
XI = qt.tensor(X,I)
YI = qt.tensor(Y,I)

zero = qt.basis(2,0)
one = qt.basis(2,1)
plus = (zero+one).unit()
minus = (zero-one).unit()
plus_i = (zero+1j*one).unit()
minus_i = (zero-1j*one).unit()

X90 = qt.operations.rx(np.pi/2)
Y90 = qt.operations.ry(np.pi/2)
mX90 = qt.operations.rx(-np.pi/2)
mY90 = qt.operations.ry(-np.pi/2)
X180 = qt.operations.rx(np.pi)
Y180 = qt.operations.ry(np.pi)
Z180 = qt.operations.rz(np.pi)
CX90 = (
    qt.operations.controlled_gate(X90, control=1, target=0, control_value=0) 
    * qt.operations.controlled_gate(mX90, control=1, target=0, control_value=1))


proj_0 = qt.projection(2, 0, 0)
proj_1 = qt.projection(2, 1, 1)

#BSM 1 (on the middle node, Bob)
# 0 - communication alice, 1 - memory bob, 2 - communication bob, 3 - communication charlie
ent_gate = _circuit_list_to_propagator([qt.tensor([I, mX90]), CX90, qt.tensor([I, X90])])
tensor_ent_gate_BSM1 = qt.tensor([I, ent_gate, I])

result_00_BSM1 = qt.tensor([I, proj_0, proj_0, I])
result_01_BSM1 = qt.tensor([I, proj_0, proj_1, I])
result_10_BSM1 = qt.tensor([I, proj_1, proj_0, I])
result_11_BSM1 = qt.tensor([I, proj_1, proj_1, I])

result_dict_BSM1 = {
    '00': result_00_BSM1,
    '01': result_01_BSM1,
    '10': result_10_BSM1,
    '11': result_11_BSM1,
}

feedback_00_BSM1 = qt.tensor([I, I, I, X180*mY90])
feedback_01_BSM1 = qt.tensor([I, I, I, X180*Y90])
feedback_10_BSM1 = qt.tensor([I, I, I, Y180*mY90])
feedback_11_BSM1 = qt.tensor([I, I, I, Y180*Y90])

feedback_dict_BSM1 = {
    '00': feedback_00_BSM1,
    '01': feedback_01_BSM1,
    '10': feedback_10_BSM1,
    '11': feedback_11_BSM1,
}

#BSM 2 (on Charlie)
# 0 - communication alice, 1 - memory charlie, 2 - communication charlie
ent_gate = _circuit_list_to_propagator([qt.tensor([I, mX90]), CX90, qt.tensor([I, X90])])
tensor_ent_gate_BSM2 = qt.tensor([I, ent_gate])


result_00_BSM2 = qt.tensor([I, proj_0, proj_0])
result_01_BSM2 = qt.tensor([I, proj_0, proj_1])
result_10_BSM2 = qt.tensor([I, proj_1, proj_0])
result_11_BSM2 = qt.tensor([I, proj_1, proj_1])

result_dict_BSM2 = {
    '00': result_00_BSM2,
    '01': result_01_BSM2,
    '10': result_10_BSM2,
    '11': result_11_BSM2,
}

feedback_11_BSM2 = qt.tensor([X180*mY90, I, I])
feedback_10_BSM2 = qt.tensor([X180*Y90, I, I])
feedback_01_BSM2 = qt.tensor([Y180*mY90, I, I])
feedback_00_BSM2 = qt.tensor([Y180*Y90, I, I])

feedback_dict_BSM2 = {
    '00': feedback_00_BSM2,
    '01': feedback_01_BSM2,
    '10': feedback_10_BSM2,
    '11': feedback_11_BSM2,
}


GHZ = (qt.tensor([qt.basis(2,0)]*3) + qt.tensor([qt.basis(2,1)]*3))/np.sqrt(2)
unc_GHZ = (qt.ket2dm(qt.tensor([qt.basis(2,0)]*3)) + qt.ket2dm(qt.tensor([qt.basis(2,1)]*3))) / 2

bell_ideal = qt.ket2dm(qt.bell_state('00'))

def _dephase(rho, p, target = 1):
    if target == 1:
        return (1-p)*rho + IZ*rho*IZ * p
    if target == 0:
        return (1-p)*rho + ZI*rho*ZI * p
        

def _depolarise(rho, p, target = 1):
	if target == 1:
	    return (1-p)*rho + (IZ*rho*IZ + IX*rho*IX + IY*rho*IY) * p/3.
	if target == 0:
	    return (1-p)*rho + (ZI*rho*ZI + XI*rho*XI + YI*rho*YI) * p/3.
    

def _depolarise_single_qubit(single_qubit_rho, p):
    return (1-p)*single_qubit_rho + (Z*single_qubit_rho*Z + X*single_qubit_rho*X + Y*single_qubit_rho*Y) * p/3.

def sce_state(alpha_1, alpha_2, pars):
    assert 0 < alpha_1 < 1, f"alpha_1 = {alpha_1} not in range"
    assert 0 < alpha_2 < 1, f"alpha_2 = {alpha_2} not in range"
    
    Pdc = pars["dc_rate"] * pars["det_win"]   # Dark count prob in window
    
    Puu = alpha_1 * alpha_2 * (
        (1 - Pdc)**2 * (1-pars["pdet_psb"]) *(pars["pdet_1"] * (1 - pars["pdet_2"]) + pars["pdet_2"] * (1 - pars["pdet_1"])) 
        + 2*(1 - Pdc) * Pdc * (1-pars["pdet_psb"])**2 * (1 - pars["pdet_1"]) * (1 - pars["pdet_2"])
    )
    Pud = alpha_1 * (1 - alpha_2) * ((1 - Pdc)**2 * pars["pdet_1"] + 2*Pdc * (1 - Pdc) * (1 - pars["pdet_1"]))
    Pdu = (1 - alpha_1) * alpha_2 * ((1 - Pdc)**2 * pars["pdet_2"] + 2*Pdc * (1 - Pdc) * (1 - pars["pdet_2"]))
    Pdd = 2 * (1 - alpha_1) * (1 - alpha_2) * Pdc * (1 - Pdc)
    
    Pe = 0.5 * (1 - np.exp(-0.5 * (pars["phase_u"]/180*np.pi)**2)) # Phase uncertainty
    
    Ptot = Puu + Pud + Pdu + Pdd
    rate = Ptot / pars["lde_length"]
    fidelity = (1-pars["depol"]) * (Pud + Pdu + 2 * (1 - pars["p2e"])**2 * (1 - 2*Pe) * np.sqrt(pars["vis"] * Pud * Pdu)) / (2*Ptot) + pars["depol"]/4
    
    rho = qt.Qobj(np.array([
            [Pdd, 0, 0, 0],
            [0, Pud, np.sqrt(pars["vis"] * Pud * Pdu), 0],
            [0, np.sqrt(pars["vis"] * Pud * Pdu), Pdu, 0],
            [0, 0, 0, Puu]]) / Ptot, dims=[[2, 2], [2, 2]])

    # Simulate double excitation
    rho = ((1 - pars["p2e"]/2) ** 2 * rho \
          + (1-pars["pdet_psb"])*(1 - pars["p2e"]/2) * (pars["p2e"]/2) * ZI * rho * ZI \
          + (1-pars["pdet_psb"])*(1 - pars["p2e"]/2) * (pars["p2e"]/2) * IZ * rho * IZ \
          + (1-pars["pdet_psb"])**2*(pars["p2e"]/2) ** 2 * ZZ * rho * ZZ).unit()
    
    # Phase uncertainty
    rho = (1 - Pe) * rho + Pe * IZ * rho * IZ

    # Depolarizing noise      
    rho = (1-pars["depol"])*rho + pars["depol"] * II/4
    
    return rho, rate, fidelity

def teleport_state(alpha_1_a, alpha_1_b, alpha_2_b, alpha_2_c, timeout, state_to_teleport):
    
    assert timeout >= 1, "Wrong timeout"
    timeout = int(timeout)
    
    rho_ab, rate_ab, fid_ab = sce_state(alpha_1_a, alpha_1_b, parameters["epr_ab"])
            
    # Swap operation mapping is equivalent to applying Hadamard
    rho_ab = _transform_rho(rho_ab, IH)
    
    # Depolarise and dephase memory
    rho_ab = _depolarise(rho_ab, parameters["memory_bob"]["depol"] * 0.75)
    N1e = np.interp(alpha_2_b, parameters["memory_bob"]["alphas"], parameters["memory_bob"]["N1e"])
    power = np.interp(alpha_2_b, parameters["memory_bob"]["alphas"], parameters["memory_bob"]["power"])
    
    #We average over possible memory dephasings.
    p_deph = (1 - np.exp(-(np.arange(1, timeout) / N1e)**power)).mean()
    rho_ab = _dephase(rho_ab, p_deph * 0.5)
    
    #Dephasing and depolarization due to imperfect decouplings
    p_depol = parameters["communication_alice"]["depolarizing"]
    rho_ab = _depolarise(rho_ab, p_depol * 0.75, target = 0)
    p_dephase = parameters["communication_alice"]["dephasing"]
    rho_ab = _dephase(rho_ab, p_dephase * 0.5, target = 0)
    
    rho_bc, rate_bc, fid_bc = sce_state(alpha_2_b, alpha_2_c, parameters["epr_bc"])
    p_depol_b = parameters["communication_bob"]["depolarizing"]
    p_depol_c = parameters["communication_charlie"]["depolarizing"]
    rho_bc = _depolarise(rho_bc, p_depol_b * 0.75, target = 0)
    rho_bc = _depolarise(rho_bc, p_depol_c * 0.75, target = 1)

    p_dephase_b = parameters["communication_bob"]["dephasing"]
    p_dephase_c = parameters["communication_charlie"]["dephasing"]
    rho_bc = _dephase(rho_bc, p_dephase_b * 0.5, target = 0)
    rho_bc = _dephase(rho_bc, p_dephase_c * 0.5, target = 1)
        
    # 0 - communication alice, 1 - memory bob, 2 - communication bob, 3 - communication charlie
    full_rho = qt.tensor([rho_ab, rho_bc])
    
    # Apply entangling gate
    full_rho = _transform_rho(full_rho, tensor_ent_gate_BSM1)
    
    # Bell state measurement middle node
    wrong_ro = {
        "0": "1",
        "1": "0"
    }
    
    swapped_rho = 0
    
    ro_fids_c = parameters["bob_ro_fidelities"]['memory']
    ro_fids_e = parameters["bob_ro_fidelities"]['communication']
    BSM_result = parameters["BSM"]["bob"]
    for r_c in ["0", "1"]:
        for r_e in ["0", "1"]:
            result_proj = result_dict_BSM1[r_c+r_e]
            # Both correct
            if ((BSM_result=="all") or (r_c+r_e in BSM_result)):
                swapped_rho += ro_fids_c[r_c]*ro_fids_e[r_e]*_transform_rho(full_rho, _circuit_list_to_propagator([result_proj, feedback_dict_BSM1[r_c+r_e]]))
            # r_c correct, r_e wrong
            if ((BSM_result=="all") or (r_c+wrong_ro[r_e] in BSM_result)):
                swapped_rho += ro_fids_c[r_c]*(1-ro_fids_e[r_e])*_transform_rho(full_rho, _circuit_list_to_propagator([result_proj, feedback_dict_BSM1[r_c+wrong_ro[r_e]]]))
            # r_c wrong, r_e correct
            if ((BSM_result=="all") or (wrong_ro[r_c]+r_e in BSM_result)):
                swapped_rho += (1-ro_fids_c[r_c])*ro_fids_e[r_e]*_transform_rho(full_rho, _circuit_list_to_propagator([result_proj, feedback_dict_BSM1[wrong_ro[r_c]+r_e]]))
            # both wrong
            if ((BSM_result=="all") or (wrong_ro[r_c]+wrong_ro[r_e] in BSM_result)):
                swapped_rho += (1-ro_fids_c[r_c])*(1-ro_fids_e[r_e])*_transform_rho(full_rho, _circuit_list_to_propagator([result_proj, feedback_dict_BSM1[wrong_ro[r_c]+wrong_ro[r_e]]]))
    
    rho_ac = swapped_rho.unit().ptrace([0, 3])#.unit()
    
    fidelity_ac = _actual_fidelity(rho_ac,bell_ideal)
    
    # prepare state to teleport
    # rho_state_to_teleport = qt.ket2dm(state_to_teleport)
    # rho_state_to_teleport = _depolarise_single_qubit(rho_state_to_teleport, parameters["memory_charlie"]["depol"] * 0.75)
    # full_rho = qt.tensor(rho_ac, rho_state_to_teleport)

    # SWAP Alice-Charlie entangled state from communication qubit to memory, induces depolarization
    rho_ac = _depolarise(rho_ac, parameters["memory_charlie"]["depol"] * 0.75, target=1)

    # prepare state to teleport
    rho_state_to_teleport = qt.ket2dm(state_to_teleport)
    full_rho = qt.tensor(rho_ac, rho_state_to_teleport)
    
    # reordering qubits, new ordering:
    # 0 -communication alice, 1 - memory charlie, 2 - communication charlie
    full_rho = full_rho.permute([0,2,1])
    
    # entangle with swapped EPR
    full_rho = _transform_rho(full_rho, tensor_ent_gate_BSM2)
    
    teleportated_state = 0

    ideal = _transform_rho(qt.ket2dm(state_to_teleport), mY90)
    
    # Bell state measurement + feedback
    ro_fids_c = parameters["charlie_ro_fidelities"]['memory']
    ro_fids_e = parameters["charlie_ro_fidelities"]['communication']
    BSM_result = parameters["BSM"]["charlie"]
    for r_c in ["0", "1"]:
        for r_e in ["0", "1"]:
            result_proj = result_dict_BSM2[r_c+r_e]
            # Both correct
            if ((BSM_result=="all") or (r_c+r_e in BSM_result)):
                teleportated_state += ro_fids_c[r_c]*ro_fids_e[r_e]*_transform_rho(full_rho, _circuit_list_to_propagator([result_proj, feedback_dict_BSM2[r_c+r_e]]))
            # r_c correct, r_e wrong
            if ((BSM_result=="all") or (r_c+wrong_ro[r_e] in BSM_result)):
                teleportated_state += ro_fids_c[r_c]*(1-ro_fids_e[r_e])*_transform_rho(full_rho, _circuit_list_to_propagator([result_proj, feedback_dict_BSM2[r_c+wrong_ro[r_e]]]))
            # r_c wrong, r_e correct
            if ((BSM_result=="all") or (wrong_ro[r_c]+r_e in BSM_result)):
                teleportated_state += (1-ro_fids_c[r_c])*ro_fids_e[r_e]*_transform_rho(full_rho, _circuit_list_to_propagator([result_proj, feedback_dict_BSM2[wrong_ro[r_c]+r_e]]))
            # both wrong
            if ((BSM_result=="all") or (wrong_ro[r_c]+wrong_ro[r_e] in BSM_result)):
                teleportated_state += (1-ro_fids_c[r_c])*(1-ro_fids_e[r_e])*_transform_rho(full_rho, _circuit_list_to_propagator([result_proj, feedback_dict_BSM2[wrong_ro[r_c]+wrong_ro[r_e]]]))

    
    teleportated_state_rho = teleportated_state.unit().ptrace([0])


    
    #final operation to map to same basis definition
    teleportated_state_rho = _transform_rho(teleportated_state_rho, Y90)
    
    p_succ_before_timeout = 1 - (1 - rate_bc * parameters["epr_bc"]["lde_length"])**timeout
    rate_teleportation = p_succ_before_timeout / (1/(rate_ab*1) + 1/(rate_bc*1))
    fid_teleported_state = _actual_fidelity(teleportated_state_rho, qt.ket2dm(state_to_teleport))


    # Ionization
    fid_teleported_state = fid_teleported_state*(1-parameters["communication_alice"]["ionization"])
    
    return teleportated_state_rho,rho_ac,fid_ab,fid_bc, rate_teleportation, fidelity_ac,fid_teleported_state