# Qubit teleportation between non-neighboring nodes in a quantum network
# Hermans, Pompili, et al.

# Simulations for basis-alternating repetitive readout
# Note: c (carbon nuclear spin) refers to memory qubit, e (electronic spin) refers to communication qubit

import qutip as qt
import numpy as np
from tqdm import tqdm

parameters = {
	'Bob' : {
		'init_state_fidelity' : 0.98,
		'F0' : 0.92,
		'F1' : 0.995,

		'depol_cond_gates' :  0.021,
		'depol_unc_gate' :  0.021/2.,
		'depol_ms0_RO' :  0.011,
		},

	'Charlie' : {
		'init_state_fidelity' : 0.98,
		'F0' : 0.928,
		'F1' : 0.99,

		'depol_cond_gates' :  0.035,
		'depol_unc_gate' :  0.035/2.,
		'depol_ms0_RO' :  0.015,
	},
}

def _actual_fidelity(x, y):
    return qt.fidelity(x, y)**2

def _circuit_list_to_propagator(circuit_list):
    propagator = circuit_list[-1]
    for g in reversed(circuit_list[:-1]):
        propagator *= g
    return propagator

def _transform_rho(rho, gate):
    return gate * rho * gate.dag()

I = qt.identity(2)
X = qt.sigmax()
Y = qt.sigmay()
Z = qt.sigmaz()
H = qt.operations.hadamard_transform(1)
II = qt.identity([2, 2])
ZI = qt.tensor(Z, I)
IZ = qt.tensor(I, Z)
XX = qt.tensor(X, X)
YY = qt.tensor(Y, Y)
ZZ = qt.tensor(Z, Z)
IH = qt.tensor(I, H)
IX = qt.tensor(I, X)
IY = qt.tensor(I, Y)
XI = qt.tensor(X,I)
YI = qt.tensor(Y,I)

zero = qt.basis(2,0)
one = qt.basis(2,1)
plus = (zero+one).unit()
minus = (zero-one).unit()
plus_i = (zero+1j*one).unit()
minus_i = (zero-1j*one).unit()

X90 = qt.operations.rx(np.pi/2)
Y90 = qt.operations.ry(np.pi/2)
mX90 = qt.operations.rx(-np.pi/2)
mY90 = qt.operations.ry(-np.pi/2)
X180 = qt.operations.rx(np.pi)
Y180 = qt.operations.ry(np.pi)
Z180 = qt.operations.rz(np.pi)
CX90 = (
    qt.operations.controlled_gate(X90, control=1, target=0, control_value=0) 
    * qt.operations.controlled_gate(mX90, control=1, target=0, control_value=1))
CY90 = (
    qt.operations.controlled_gate(Y90, control=1, target=0, control_value=0) 
    * qt.operations.controlled_gate(mY90, control=1, target=0, control_value=1))
CmX90 = (
    qt.operations.controlled_gate(mX90, control=1, target=0, control_value=0) 
    * qt.operations.controlled_gate(X90, control=1, target=0, control_value=1))
CmY90 = (
    qt.operations.controlled_gate(mY90, control=1, target=0, control_value=0) 
    * qt.operations.controlled_gate(Y90, control=1, target=0, control_value=1))

proj_0 = qt.projection(2, 0, 0)
proj_1 = qt.projection(2, 1, 1)


def _dephase(rho, p, target = 1):
    if target == 1:
        return (1-p)*rho + IZ*rho*IZ * p
    if target == 0:
        return (1-p)*rho + ZI*rho*ZI * p
        
def _depolarise(rho, p):
    return (1-p)*rho + (ZI*rho*ZI + XI*rho*XI + YI*rho*YI) * p/3.

def _depolarise_single_qubit(single_qubit_rho, p):
    return (1-p)*single_qubit_rho + (Z*single_qubit_rho*Z + X*single_qubit_rho*X + Y*single_qubit_rho*Y) * p/3.

def single_readout(c_state,Z,depol_cond_gates,depol_unc_gate,depol_ms0_RO,F0, F1):
    # Initializing electron state
    e_state = qt.ket2dm(zero)
    rho = qt.tensor(c_state,e_state)

    # Conditional gate to map c_state to electron (Z readout)
    # depolarisation due to conditional gates
    rho = _depolarise(rho, depol_cond_gates*0.75)
    if Z:
        cond_gates = _circuit_list_to_propagator([CX90,qt.tensor(I,Y90),CmY90,qt.tensor(I,X90)])
    else:
        cond_gates = _circuit_list_to_propagator([CX90,qt.tensor(I,Y90),CY90,qt.tensor(I,X90)])
    rho = _transform_rho(rho, cond_gates)

    # Unconditional gate
    rho = _depolarise(rho, depol_unc_gate*0.75)
    rho = _transform_rho(rho, qt.tensor(mX90,I))

    # Electron readout + depolarization of the memory qubit due to readout
    p0 = (proj_0*rho.ptrace([1])*proj_0.dag()).tr()
    true_outcome = np.random.binomial(1, 1-p0)

    if true_outcome == 0:
        c_state = _depolarise_single_qubit(c_state, depol_ms0_RO*0.75) 
        rho = _transform_rho(rho,qt.tensor(I,proj_0))
        c_state = rho.ptrace([0]).unit()
        measured_outcome = np.random.binomial(1,1-F0)
    elif true_outcome == 1:
        rho = _transform_rho(rho,qt.tensor(I,proj_1))
        c_state = rho.ptrace([0]).unit()
        measured_outcome = np.random.binomial(1,F1)
        
    return c_state, true_outcome, measured_outcome

def repetitive_readout(nof_readouts = 5, nof_msmts = int(2e3),init_state_fidelity = 1,
	depol_cond_gates = 0,
	depol_unc_gate = 0,
	depol_ms0_RO = 0,
	F0 = 1, 
	F1 = 1):
	init_states = [np.sqrt(init_state_fidelity)*zero + np.sqrt(1-init_state_fidelity)*one, 
					np.sqrt(init_state_fidelity)*one + np.sqrt(1-init_state_fidelity)*zero]
	p0 = init_state_fidelity
	p1 = init_state_fidelity


	c_state_after_RO = []
	true = np.zeros((2,nof_msmts, nof_readouts),dtype=int)
	measured = np.zeros((2,nof_msmts, nof_readouts),dtype=int)

	for k, init_state in enumerate(init_states):
	    for i in tqdm(range(nof_msmts)):
	        c_state = qt.ket2dm(init_state)
	        for j in range(nof_readouts):
	            result = single_readout(c_state,bool((j+1)%2),depol_cond_gates,depol_unc_gate,depol_ms0_RO, F0, F1)
	            c_state = result[0]
	            c_state_after_RO.append(c_state)
	            true[k,i,j] = result[1]
	            measured[k,i,j] = result[2]

	# flipping back the even readouts (easier for further analysis)
	measured[:,:,1::2] = (measured[:,:,1::2]+1)%2
	true[:,:,1::2] = (true[:,:,1::2]+1)%2

	nof_zeros = np.zeros((2,nof_readouts))
	nof_ones = np.zeros((2,nof_readouts))
	F = np.zeros_like(nof_zeros)
	rejected_fraction = np.zeros_like(nof_zeros)
	m0 = np.zeros(nof_readouts)
	m1 = np.zeros(nof_readouts)

	for k in range(len(init_states)):
	    for j in range(nof_readouts):
	        nof_zeros[k,j] = np.sum(np.all(measured[k,:,:(j+1)]==0 ,axis=1))
	        nof_ones[k,j] = np.sum(np.all(measured[k,:,:(j+1)]==1 ,axis=1))
	    
	    if k==0:
	        m0 = nof_zeros[k]/(nof_zeros[k]+nof_ones[k])
	    elif k==1:
	        m1 = nof_ones[k]/(nof_zeros[k]+nof_ones[k])
	    rejected_fraction[k] = 1 - (nof_zeros[k]+nof_ones[k])/nof_msmts
	F[0] = m0*((-p0 + 1)*(-p1 + 1)/(p0*(p0 + p1 - 1)) + 1/p0) - (-m1 + 1)*(-p0 + 1)/(p0 + p1 - 1)
	F[1] = m1*p0/(p0 + p1 - 1) - (-m0 + 1)*(-p1 + 1)/(p0 + p1 - 1)


	result = {}
	result = {
	    'F' : F,
	    'rejected_fraction' : rejected_fraction,
	    'true' : true,
	    'measured' : measured,
	    'init_state_fidelity' : init_state_fidelity,
	    'F0' : F0,
	    'F1' : F1,
	    'depol_cond_gates' :  depol_cond_gates,
	    'depol_unc_gate' :  depol_unc_gate,
	    'depol_ms0_RO' :  depol_ms0_RO,
	}
	    
	return result

    

